{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "aa622fbf",
   "metadata": {},
   "source": [
    "# Binary Search Trees\n",
    "\n",
    "As we mentioned in the introduction, Python does not use Binary Search Trees in any of its abstract data types.\n",
    "\n",
    "Here we include a program for the implementation that is described in the book [Algorithms Illuminated](http://www.algorithmsilluminated.org) in chapters 11.1, 11.2 and 11.3 from part 2. The website includes four videos with explanations. \n",
    "\n",
    "The implementation is borrowed from the website of another book: [Introduction to Programming in Python](https://introcs.cs.princeton.edu/python/home/) by R. Sedgewick, K. Wayne and R. Dondero, chapter 4.4.\n",
    "\n",
    "In order to understand why we used the names that we used for the methods of the class, check the comments in the little application in the cell immediately after the implementation.\n",
    "\n",
    "The workings of this (very naive) implementation can be visualized in [Binary Search Tree](https://www.cs.usfca.edu/~galles/visualization/BST.html)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1fc0e3f2",
   "metadata": {},
   "outputs": [],
   "source": [
    "# An ordered dictionary implemented as binary search trees \n",
    "# (keeps keys in order) \n",
    "# as implemented in the book Introduction to Programming in Python\n",
    "class ordered_dict:\n",
    "\n",
    "    #-------------------------------------------------------------------\n",
    "\n",
    "    # Construct a new ordered_dict object.\n",
    "\n",
    "    def __init__(self):\n",
    "        self._root = None  # Reference to root _Node object\n",
    "        \n",
    "    \n",
    "    #-------------------------------------------------------------------\n",
    "\n",
    "    # Search the subtree of self whose root is x for a _Node object\n",
    "    # with the given key.  If found, return that _Node object's value;\n",
    "    # otherwise raise a KeyError.\n",
    "\n",
    "    def _get(self, x, key):\n",
    "        if x is None:\n",
    "            raise KeyError\n",
    "        if key < x.key:\n",
    "            return self._get(x.left, key)\n",
    "        elif x.key < key:\n",
    "            return self._get(x.right, key)\n",
    "        else:\n",
    "            return x.val\n",
    "\n",
    "   # Return the value associated with key in self.\n",
    "\n",
    "    def __getitem__(self, key):\n",
    "        return self._get(self._root, key)\n",
    "\n",
    "    #-------------------------------------------------------------------\n",
    "\n",
    "    # x is the root of a subtree self.  If a _Node object with\n",
    "    # the given key exists in that subtree, then set its\n",
    "    # value to val.  Otherwise insert a new _Node object consisting\n",
    "    # of the given key and val into the subtree.  Return the root of\n",
    "    # the resulting subtree.\n",
    "\n",
    "    def _set(self, x, key, val):\n",
    "        if x is None:\n",
    "            return _Node(key, val)\n",
    "        \n",
    "        if key < x.key:\n",
    "            x.left = self._set(x.left, key, val)\n",
    "            \n",
    "        elif x.key < key:\n",
    "            x.right = self._set(x.right, key, val)\n",
    "            \n",
    "        else:\n",
    "            x.val = val\n",
    "        return x\n",
    "\n",
    "    # Associate key with val in self.\n",
    "\n",
    "    def __setitem__(self, key, val):\n",
    "        self._root = self._set(self._root, key, val)\n",
    "\n",
    "    #-------------------------------------------------------------------\n",
    "\n",
    "    # Search the subtree of self whose root is x for a _Node object\n",
    "    # with the given key.  If found, return True; otherwise return\n",
    "    # False.\n",
    "\n",
    "    def _contains(self, x, key):\n",
    "        if x is None:\n",
    "            return False\n",
    "        if key < x.key:\n",
    "            return self._contains(x.left, key)\n",
    "        if x.key < key:\n",
    "            return self._contains(x.right, key)\n",
    "        return True\n",
    "\n",
    "    # Return True if key is in self, and False otherwise.\n",
    "\n",
    "    def __contains__(self, key):\n",
    "        return self._contains(self._root, key)\n",
    "\n",
    "    #-------------------------------------------------------------------\n",
    "\n",
    "    # Populate list a with all keys in the subtree of self whose\n",
    "    # root is x.\n",
    "\n",
    "    def _inorder(self, x, a):\n",
    "        if x is None:\n",
    "            return\n",
    "        self._inorder(x.left, a)\n",
    "        a += [x.key]\n",
    "        self._inorder(x.right, a)\n",
    "\n",
    "    # Return an iterator for ordered_dict object self.\n",
    "\n",
    "    def __iter__(self):\n",
    "        a = []\n",
    "        self._inorder(self._root, a)\n",
    "        return iter(a)\n",
    "\n",
    "    \n",
    "#-----------------------------------------------------------------------\n",
    "\n",
    "# A _Node object references a key, a value, and left and right\n",
    "# children _Node objects.  An ordered_dict object is composed of\n",
    "# _Node objects.\n",
    "\n",
    "class _Node:\n",
    "    def __init__(self, key, val):\n",
    "        self.key = key    # Reference to key\n",
    "        self.val = val    # Reference to value\n",
    "        self.left = None  # Reference to left child _Node object\n",
    "        self.right = None # Reference to right child _Node object\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b0a1bd46",
   "metadata": {},
   "source": [
    "What follows is an example where we use the ordered dictionary."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "09d19eae",
   "metadata": {},
   "outputs": [],
   "source": [
    "# using ordered_dict calls __init__\n",
    "phoneNumbers = ordered_dict() \n",
    "\n",
    "\n",
    "# assigning a value to the ordered dictionary at a key uses __setitem__\n",
    "phoneNumbers['vero']   = 123  \n",
    "phoneNumbers['martin'] = 456\n",
    "phoneNumbers['tim']    = 789\n",
    "\n",
    "# accessing the ordered dictionary at a given key uses __getitem__\n",
    "print(phoneNumbers['martin'])\n",
    "\n",
    "# testing whether a key is in the ordered dictionary using in uses __contains__\n",
    "if  'anna' in phoneNumbers: \n",
    "    print(phoneNumbers['anna'])\n",
    "else: \n",
    "    phoneNumbers['anna'] = phoneNumbers['tim']\n",
    "    print(phoneNumbers['anna'])\n",
    "\n",
    "# iterating over the keys uses __iter__\n",
    "for key in phoneNumbers:\n",
    "    print (key, phoneNumbers[key])"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a99eb921",
   "metadata": {},
   "source": [
    "If you run this code you will see that the contents of the dictionary is printed in order (of the keys). No matter in what order the items were added, they appear sorted.\n",
    "\n",
    "It is enough to change the first line so that instead of creating an ```ordered_dict```you create a ```dict``` to see that the output is **not** ordered: hash tables (the data structure used to implement ```dict```) does not keep an order among the items in the collection.\n",
    "\n",
    "## Counting words (again!)\n",
    "\n",
    "We do the same trick of modifying the program for counting the words in a file to use the binary search tree implementation of a dictionary (```ordered_dict```) and check the output:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f9ecae66",
   "metadata": {},
   "outputs": [],
   "source": [
    "# This is the same program as word_count but using an \n",
    "# ordered_dict instead of a dict.\n",
    "# As you can see the resulting output file is ordered!\n",
    "\n",
    "import re\n",
    "def word_count_sorted(input_file, output_file):\n",
    " \n",
    "    f = open(input_file, 'r')\n",
    "    txt  = f.read()\n",
    "    f.close()\n",
    "\n",
    "    text = re.findall(r\"[a-zA-Z']+\", txt)\n",
    "    \n",
    "    # Now to the algorithm: just lookup and update words in the dict\n",
    "    # (implemented with a binary search tree.\n",
    "    # In our example the keys are words and the values are the count.\n",
    "    \n",
    "    counts = ordered_dict()\n",
    "    \n",
    "    for word in text:\n",
    "        w = word.lower()\n",
    "        if w in counts: \n",
    "            counts[w] += 1\n",
    "        else:\n",
    "            counts[w] = 1\n",
    "    \n",
    "    f = open(output_file,'w')\n",
    "    \n",
    "    # write one (word, count) at a time\n",
    "    len = 0\n",
    "    for word in counts:\n",
    "        f.write(word + ' '+ str(counts[word]) + '\\n')\n",
    "        len +=1\n",
    "\n",
    "    f.close()\n",
    "    return len"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8ff19ba2",
   "metadata": {},
   "outputs": [],
   "source": [
    "word_count_sorted('tomsawyer.txt', 'tomsawyer.counts')"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c44c51f2",
   "metadata": {},
   "source": [
    "### Quiz\n",
    "\n",
    "What is the Big-O for your program? Can you argue for it? \n",
    "\n",
    "### Exercise\n",
    "Modify the implementation of ordered_dict \n",
    "by adding a function min that returns the (key,value)\n",
    "for the minimum key. What is the Big-O for the execution time?\n",
    "\n",
    "### Exercise\n",
    "Modify the implementation of ordered_dict \n",
    "by adding a function max that returns the (key,value)\n",
    "for the maximum key. What is the Big-O for the execution time?\n",
    "\n",
    "### Exercise\n",
    "Modify the implementation of ordered_dict \n",
    "by adding a function height that returns the height of the tree.\n",
    "Use this method to calculate the height of the tree constructed to count the words in Tom Sawyer. \n",
    "\n",
    "**The height of a tree is the number of edges from the root to the deepest leaf**. The height of a binary search tree with only one node is 0. In the binary search tree \n",
    "\n",
    "<img src=\"height.png\" width=\"100\"/>\n",
    "\n",
    "the height is 2: the number of edges from the root (5) to the deepest leaf (4).\n",
    "\n",
    "\n",
    "\n",
    "### Exercise\n",
    "Write a program that creates an ordered dictionary \n",
    "and inserts the elements of range(10000) in order. \n",
    "What is the height of the height of the tree? \n",
    "\n",
    "### Exercise\n",
    "\n",
    "What could you do if you wanted the output to be sorted by the count instead of by the word? What is the Big-O of your solution?"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
